<?xml version="1.0"?>
<!DOCTYPE gui SYSTEM "kpartgui.dtd">

<gui name="konsole" version="20">
    <MenuBar>
        <Menu name="file"><text>File</text>
            <Action name="new-window"/>
            <Action name="new-tab"/>
            <Action name="clone-tab"/>
            <Separator/>
            <DefineGroup name="session-operations"/>
            <Separator/>
            <DefineGroup name="session-tab-operations"/>
            <Action name="close-window"/>
        </Menu>
        <Menu name="edit"><text>Edit</text>
            <DefineGroup name="session-edit-operations"/>
        </Menu>
        <Menu name="view"><text>View</text>
            <Menu name="view-split"><text>Split View</text>
                <Action name="split-view-left-right"/>
                <Action name="split-view-top-bottom"/>
                <Action name="split-view-left-right-next-tab"/>
                <Action name="split-view-top-bottom-next-tab"/>
                <Action name="close-active-view"/>
                <Action name="close-other-views"/>
                <Action name="expand-active-view"/>
                <Action name="shrink-active-view"/>
                <Action name="toggle-maximize-current-view"/>
                <Action name="equal-size-view"/>
            </Menu>
            <Separator/>
            <Action name="detach-tab" />
            <Action name="detach-view"/>
            <Action name="save-layout"/>
            <Action name="load-layout"/>
            <Separator/>
            <DefineGroup name="session-view-operations"/>
        </Menu>
        <Action name="bookmark"/>
        <Menu name="settings"><text>Settings</text>
            <DefineGroup name="session-settings"/>
            <Action name="manage-profiles"/>
            <Action name="show-menubar"/>
            <Action name="window-colorscheme-menu"/>
            <Separator/>
            <Action name="view-full-screen"/>
            <Separator/>
            <Action name="configure-shortcuts"/>
            <Action name="configure-notifications"/>
            <Action name="configure-settings"/>
        </Menu>
        <Menu name="plugins">
            <text>Plugins</text>
            <ActionList name="plugin-submenu" />
        </Menu>
        <Menu name="help"><text>Help</text>
        </Menu>
    </MenuBar>
    <ToolBar name="mainToolBar">
        <text>Main Toolbar</text>
        <index>0</index>
        <Action name="new-tab"/>
        <Action name="split-view"/>
    </ToolBar>
</gui>
