/*
    SPDX-FileCopyrightText: 2019 Tomaz Canabrava <tomaz.canabrava@kde.org>

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "CharacterTest.h"
#include "Character.h"

#include <QTest>
#include <cstdint>

QTEST_GUILESS_MAIN(Konsole::CharacterTest)

#include "moc_CharacterTest.cpp"
