/*
    SPDX-FileCopyrightText: 2023 Waqar Ahmed <waqar.17a@gmail.com>
    SPDX-License-Identifier: MIT
*/
#include "iptyprocess.h"

IPtyProcess::~IPtyProcess() = default;

#include "moc_iptyprocess.cpp"
